@@ -12,6 +12,7 @@ from random import getrandbits, randrange
 from typing import Any, Callable, Coroutine, Optional, Tuple, TypeVar, Union, cast, List, Dict
 from textwrap import wrap
 from os import path
+from hashlib import sha1
 
 from .constants import *
 from .logger.constants import *
@@ -546,8 +547,6 @@ class Bus(EventEmitter):
                     matches = False
                     if not dev:
                         dev = Device(self, pkt.device_id, pkt.data)
-                        # ask for uptime
-                        # dev.send_ctrl_command(CMD_GET_REG | ControlReg.Uptime)
                         self.emit(EV_DEVICE_CONNECT, dev)
                     else:
                         matches = _service_matches(dev, pkt.data)
@@ -990,6 +989,7 @@ class ControlServer(Server):
     def __init__(self, bus: Bus) -> None:
         super().__init__(bus, JD_SERVICE_CLASS_CONTROL)
         self.restart_counter = 0
+        self.auto_bind_cnt = 0
 
     def queue_announce(self):
         logv("announce: %d " % self.restart_counter)
@@ -1009,12 +1009,12 @@ class ControlServer(Server):
         self.send_report(JDPacket(cmd=0, data=buf))
 
         # auto bind
-        # if jacdac.role_manager_server.auto_bind:
-        #     self.auto_bind_cnt++
-        #     # also, only do it every two announces (TBD)
-        #     if self.auto_bind_cnt >= 2:
-        #         self.auto_bind_cnt = 0
-        #         jacdac.role_manager_server.bind_roles()
+        if self.bus.role_manager and self.bus.role_manager.auto_bind:
+            self.auto_bind_cnt += 1
+            # also, only do it every two announces (TBD)
+            if self.auto_bind_cnt >= 2:
+                self.auto_bind_cnt = 0
+                self.bus.role_manager.bind_roles()
 
     # def handle_flood_ping(self, pkt: JDPacket):
     #     num_responses, counter, size = pkt.unpack("IIB")
@@ -1126,7 +1126,99 @@ class UniqueBrainServer(Server):
         super().__init__(bus, JD_SERVICE_CLASS_UNIQUE_BRAIN)
 
 
+class DeviceWrapper:
+    def __init__(self, device: 'Device') -> None:
+        self.device = device
+        self.bindings: Dict[int, 'RoleBinding'] = {}
+        self.score = -1
+
+
+class RoleBinding:
+    def __init__(self, role_manager: 'RoleManagerServer', role: str, service_class: int) -> None:
+        self.role_manager = role_manager
+        self.role = role
+        self.service_class = service_class
+        self.bound_to_dev: Optional[Device] = None
+        self.bound_to_service_idx: Optional[int] = None
+
+    def host(self) -> str:
+        slash_idx = self.role.find("/")
+        if slash_idx < 0:
+            return self.role
+        else:
+            return self.role[0: slash_idx]
+
+    def select(self, devwrap: DeviceWrapper, service_idx: int):
+        dev = devwrap.device
+        if dev == self.bound_to_dev and service_idx == self.bound_to_service_idx:
+            return
+
+        devwrap.bindings[service_idx] = self
+        self.role_manager.set_role(self.role, dev, service_idx)
+        self.bound_to_dev = dev
+        self.bound_to_service_idx = service_idx
+
+
+class ServerBindings:
+    def __init__(self, host: str) -> None:
+        self.host = host
+        self.bindings: List[RoleBinding] = []
+
+    @property
+    def fully_bound(self) -> bool:
+        for binding in self.bindings:
+            if not binding.bound_to_dev:
+                return False
+        return True
+
+    def score_for(self, devwrap: DeviceWrapper, select: Optional[bool] = False):
+        """candidate devices are ordered by [numBound, numPossible, device_id]
+        where numBound is number of clients already bound to this device
+        and numPossible is number of clients that can possibly be additionally bound
+        """
+        num_bound = 0
+        num_possible = 0
+        dev = devwrap.device
+        missing: List[RoleBinding] = []
+        for b in self.bindings:
+            if b.bound_to_dev:
+                if b.bound_to_dev == dev:
+                    num_bound += 1
+            else:
+                missing.append(b)
+
+        sbuf = dev.services
+        for idx in range(4, len(sbuf), 4):
+            service_index = idx >> 2
+            # if service is already bound to some client, move on
+            if service_index in devwrap.bindings:
+                continue
+
+            service_class = util.u32(sbuf, idx)
+            for i in range(len(missing)):
+                if missing[i].service_class == service_class:
+                    # we've got a match!
+                    num_possible += 1  # this can be assigned
+                    # in fact, assign if requested
+                    if select:
+                        missing[i].select(devwrap, service_index)
+                    # this one is no longer missing
+                    missing.pop(i)
+                    # move on to the next service in announce
+                    break
+
+        # if nothing can be assigned, the score is zero
+        if num_possible == 0:
+            return 0
+
+        # otherwise the score is [numBound, numPossible], lexicographic
+        # numPossible can't be larger than ~64, leave it a few more bits
+        return (num_bound << 8) | num_possible
+
+
 class RoleManagerServer(Server):
+    """A server for the role manager service
+    """
     def __init__(self, bus: Bus) -> None:
         super().__init__(bus, JD_SERVICE_CLASS_ROLE_MANAGER)
 
@@ -1135,6 +1227,8 @@ class RoleManagerServer(Server):
         self.settings = SettingsFile(file_name)
         self.auto_bind = 1
 
+        self._old_binding_hash = ""
+
     def handle_packet(self, pkt: JDPacket):
         self.auto_bind = self.handle_reg_u8(
             pkt, JD_ROLE_MANAGER_REG_AUTO_BIND, self.auto_bind)
@@ -1172,15 +1266,10 @@ class RoleManagerServer(Server):
 
     def handle_list_stored_roles(self, pkt: JDPacket):
         pipe = OutPipe(self.bus, pkt)
-        for role in self.settings.list():
-            payload = self.settings.read(role)
+        for role, payload in self.settings.list():
             pipe.write(payload)
         pipe.close()
 
-    def handle_clear_all_roles(self, pkt: JDPacket):
-        self.settings.clear()
-        self.send_change_event()
-
     def handle_get_role(self, pkt: JDPacket):
         payload = pkt.unpack("b[8] u8")
         device_id_b = cast(bytearray, payload[0])
@@ -1195,18 +1284,132 @@ class RoleManagerServer(Server):
         self.send_report(JDPacket.packed(
             JD_ROLE_MANAGER_CMD_GET_ROLE, "u[8] u8 s", device_id_b, service_index, role))
 
+    def handle_all_roles_allocated(self, pkt: JDPacket):
+        res = 1
+        for client in self.bus.all_clients:
+            if not client.broadcast and not client.device:
+                res = 0
+                break
+        self.send_report(JDPacket.packed(pkt.service_command, "u8", res))
+
+    def handle_clear_all_roles(self, pkt: JDPacket):
+        self.settings.clear()
+        self.bus.clear_attach_cache()
+        self.bind_roles()
+
     def handle_set_role(self, pkt: JDPacket):
         payload = pkt.unpack("b[8] u8 s")
-        device_id_b = cast(bytearray, payload[0])
-        device_id = device_id_b.hex()
-        service_index = cast(int, payload[1])
         role = cast(str, payload[2])
-        self.send_change_event()
-        # TODO
+        if role:
+            self.settings.write(role, pkt.data)
+            self.bus.clear_attach_cache()
+            self.bind_roles()
 
-    def handle_all_roles_allocated(self, pkt: JDPacket):
-        res = 0 if len(self.bus.unattached_clients) > 0 else 1
-        self.send_report(JDPacket.packed(pkt.service_command, "u8", res))
+    def set_role(self, role: str, device: Optional['Device'], service_idx: Optional[int]):
+        if not device or service_idx is None:
+            self.settings.delete(role)
+        else:
+            self.settings.write(role, bytearray(
+                jdpack("b[8] u8 s", bytearray.fromhex(device.device_id), service_idx, role)))
+
+    def is_match_role(self, role: str, device: 'Device', service_idx: int):
+        current = self.settings.read(role)
+        return current == jdpack("b[8] u8 s", bytearray.fromhex(device.device_id), service_idx, role)
+
+    def _binding_hash(self):
+        r = ""
+        for client in self.bus.all_clients:
+            r += "{}:{}:{},".format(client.role,
+                                    client.broadcast or client.device, client.service_index)
+        return sha1(r.encode("utf-8")).hexdigest()
+
+    def _check_changes(self):
+        new_hash = self._binding_hash()
+        if self._old_binding_hash != new_hash:
+            self._old_binding_hash = new_hash
+            self.bus.clear_attach_cache()
+            self.send_change_event()
+
+    def bind_roles(self):
+        if len(self.bus.unattached_clients) == 0:
+            self._check_changes()
+            return
+        self.debug("bind roles, {}/{} to bind",
+                   len(self.bus.unattached_clients), len(self.bus.all_clients))
+        bindings: List[RoleBinding] = []
+        wraps: List[DeviceWrapper] = []
+        for device in self.bus.devices:
+            wraps.append(DeviceWrapper(device))
+        for cl in self.bus.all_clients:
+            if not cl.broadcast and cl.role:
+                b = RoleBinding(self, cl.role, cl.service_class)
+                if cl.device:
+                    b.bound_to_dev = cl.device
+                    b.bound_to_service_idx = cl.service_index
+                    for w in wraps:
+                        if w.device == cl.device and not cl.service_index is None:
+                            w.bindings[cl.service_index] = b
+                            break
+                bindings.append(b)
+        servers: List[ServerBindings] = []
+        # Group all clients by host
+        for b in bindings:
+            hn = b.host()
+            h: Optional[ServerBindings] = None
+            for server in servers:
+                if server.host == hn:
+                    h = server
+                    break
+            if not h:
+                h = ServerBindings(hn)
+                servers.append(h)
+            h.bindings.append(b)
+
+        # exclude hosts that have already everything bound
+        servers = list(filter(lambda h: not h.fully_bound, servers))
+
+        while len(servers) > 0:
+            # Get host with maximum number of clients (resolve ties by name)
+            # This gives priority to assignment of "more complicated" hosts, which are generally more difficult to assign
+            h = servers[0]
+            for i in range(1, len(servers)):
+                a = h
+                b = servers[i]
+                if len(a.bindings) - len(b.bindings) < 0 or b.host < a.host:
+                    h = b
+            for d in wraps:
+                d.score = h.score_for(d)
+
+            dev = wraps[0]
+            for i in range(1, len(wraps)):
+                a = dev
+                b = wraps[i]
+                if a.score - b.score < 0 or b.device.device_id < a.device.device_id:
+                    dev = b
+
+            if dev.score == 0:
+                # nothing can be assigned, on any device
+                servers.remove(h)
+                continue
+
+            # assign services in order of names - this way foo/servo1 will be assigned before foo/servo2
+            # in list of advertised services
+            h.bindings = sorted(h.bindings, key=lambda entry: entry.role)
+
+            # "recompute" score, assigning names in process
+            h.score_for(dev, True)
+
+            # if everything bound on this host, remove it from further consideration
+            if h.fully_bound:
+                servers.remove(h)
+            else:
+                # otherwise, remove bindings on the current device, to update sort order
+                # it's unclear we need this
+                h.bindings = list(
+                    filter(lambda b: b.bound_to_dev != dev.device, h.bindings))
+
+        # trigger event as needed
+        self._check_changes()
 
 
 class Client(EventEmitter):
@@ -1503,8 +1706,15 @@ class Device(EventEmitter):
     def matches_role_at(self, role: str, service_idx: int):
         if not role or role == self.device_id or role == "{}:{}".format(self.device_id, service_idx):
             return True
-        return True
-        # TODO: return jacdac._rolemgr.getRole(self.deviceId, serviceIdx) == role
+        # requires role binding
+        if role.find(":") > -1:
+            return False
+
+        role_manager = self.bus.role_manager
+        if not role_manager:
+            return True
+
+        return role_manager.is_match_role(role, self, service_idx)
 
     @ property
     def num_service_classes(self):
