@@ -1,10 +1,10 @@
 from typing import Any
-from jacdac.devtools import create_dev_tools_bus
+from ..bus import Bus
 from .client import RelayClient
 from time import sleep
 
 if __name__ == '__main__':
-    bus = create_dev_tools_bus()
+    bus = Bus()
 
     def active(data: Any):
         print("active")
